from typing import List, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


from pydantic import confloat, Field, AnyUrl
from fhirkit.primitive_datatypes import URI, Instant
from fhirkit.elements import BackboneElement, Identifier
from fhirkit.Resource import Resource

BundleType = Literal[
    "document",
    "message",
    "transaction",
    "transaction-response",
    "batch",
    "batch-response",
    "history",
    "searchset",
    "collection",
]

BundleSearchMode = Literal["match", "include", "outcome"]


class BundleLink(BackboneElement):
    relation: str
    url: URI


class BundleSearch(BackboneElement):
    mode: Optional[BundleSearchMode]
    score: confloat(ge=0, le=1)


class BundleRequest(BackboneElement):
    method: Literal["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH"]
    url: URI
    ifNoneMatch: Optional[str]
    ifModifiedSince: Optional[Instant]
    ifMatch: Optional[str]
    ifNoneExist: Optional[str]


class BundleResponse(BackboneElement):
    status: str
    location: Optional[URI]
    etag: Optional[str]
    lastModified: Optional[Instant]
    outcome: Optional[Resource]


class BundleEntry(BackboneElement):
    fullUrl: Optional[URI]
    resource: Optional[Resource]
    request: Optional[BundleRequest]
    response: Optional[BundleResponse]


class Bundle(Resource):
    resourceType = Field("Bundle", const=True)
    identifier: Optional[Identifier]
    type: BundleType
    timestamp: Optional[Instant]
    entry: List[BundleEntry]



from typing import Optional
import abc
from fhirkit.elements import URI, Code, Coding, CodeableConcept
from fhirkit.Server import AbstractFHIRServer


class AbstractFHIRTerminologyServer(AbstractFHIRServer):
    @abc.abstractmethod
    def valueset_expand(self, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def valueset_validate_code(
        self,
        url: URI,
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[URI] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_terminology_resource(
        self,
        resourceType: Optional[str],
        *,
        id: Optional[str] = None,
        url: Optional[URI] = None,
    ):
        raise NotImplementedError()



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field


from fhirkit.Resource import DomainResource


class Immunization(DomainResource):
    resourceType: Literal["Immunization"] = Field("Immunization", const=True)



from .BaseModel import BaseModel
from .Resource import Resource
from .elements import CodeableConcept, Coding, Quantity, Reference
from .CodeSystem import CodeSystem
from .ValueSet import ValueSet, SimpleValueSet
from .snomed import SCTCoding, SCTConcept, SCTFHIRTerminologyServer
from .Observation import Observation
from .Procedure import Procedure
from .Encounter import Encounter
from .DocumentReference import DocumentReference
from .DiagnosticReport import DiagnosticReport
from .Immunization import Immunization
from .Condition import Condition
from .MedicationRequest import MedicationRequest
from .MedicationStatement import MedicationStatement
from .Medication import Medication
from .MedicationAdministration import MedicationAdministration
from .CareTeam import CareTeam
from .CarePlan import CarePlan
from .Organization import Organization
from .Practitioner import Practitioner
from .Provenance import Provenance
from .SupplyDelivery import SupplyDelivery
from .PractitionerRole import PractitionerRole
from .AllergyIntolerance import AllergyIntolerance
from .ImagingStudy import ImagingStudy
from .Location import Location
from .Patient import Patient
from .Device import Device

from .SimpleFHIRStore import SimpleFHIRStore



import abc
from typing import Optional, Tuple, Union
from pydantic.error_wrappers import ValidationError
from pydantic import HttpUrl, parse_obj_as
from abc import ABC


class ExpansionError(Exception):
    pass


class ResourceNotFoundError(Exception):
    pass


class AbstractFHIRServer(ABC):
    def __init__(self, base_url: Optional[Union[str, HttpUrl]] = None) -> None:
        self._base_url = base_url

    @property
    def base_url(self):
        return self._base_url

    @abc.abstractmethod
    def get_resource(
        self,
        resourceType: Optional[str],
        *,
        id: Optional[str] = None,
        uri: Optional[str] = None,
    ):
        pass

    def __getitem__(self, key: Union[str, Tuple[str, str], HttpUrl]):
        # TODO implement URL keys
        if isinstance(key, HttpUrl):
            return self.get_resource(None, uri=key)
        try:
            uri = parse_obj_as(HttpUrl, key)
        except ValidationError:
            pass
        else:
            return self.get_resource(None, uri=uri)
        if isinstance(key, str):
            if len(key.split("/")) == 2:
                resourceType, id = key.split("/")
            else:
                resourceType = None
                id = key
        elif isinstance(key, tuple):
            resourceType, id = key
        else:
            raise TypeError(f"Received a key(={key}) of type {type(key)}")
        return self.get_resource(resourceType, id=id)



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import Optional, Union
from pydantic import Field, validator
from fhirkit.choice_type import deterimine_choice_type
from fhirkit.Resource import DomainResource
from fhirkit.elements import CodeableConcept, Identifier, Period, Reference
from fhirkit.primitive_datatypes import dateTime


ProcedureStatus = Literal[
    "preparation",
    "in-progress",
    "not-done",
    "on-hold",
    "stopped",
    "completed",
    "entered-in-error",
    "unknown",
]


class Procedure(DomainResource):
    resourceType: Literal["Procedure"] = Field("Procedure", const=True)
    identifier: Optional[Identifier] = Field(None, repr=True)
    status: ProcedureStatus = Field("completed", repr=True)
    statusReason: Optional[CodeableConcept] = Field(None, repr=True)
    category: Optional[CodeableConcept] = Field(None, repr=True)
    code: Optional[CodeableConcept] = Field(None, repr=True)
    subject: Reference
    encounter: Optional[Reference] = Field(None, repr=True)
    performedDateTime: Optional[dateTime] = Field(None, exclude=True)
    performedPeriod: Optional[Period] = Field(None, exclude=True)
    performedString: Optional[str] = Field(None, exclude=True)
    performed: Union[dateTime, Period, str] = Field(None, repr=True)

    @validator("performed", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, v, values, field):
        return deterimine_choice_type(
            cls,
            v,
            values,
            field,
        )



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from typing import Any, List, Optional, Union
from pydantic import (
    Field,
    HttpUrl,
    StrictInt,
    StrictStr,
    StrictBool,
    validator,
)
from fhirkit.choice_type import deterimine_choice_type
from fhirkit.Resource import Resource
from fhirkit.primitive_datatypes import Code
from fhirkit.elements import BackboneElement, CodeableConcept, Coding


class AbstractParameter(BackboneElement):
    name: str


class ValueParameter(AbstractParameter):
    valueBoolean: Optional[bool] = Field(None, exclude=True)
    value: Union[
        StrictBool, StrictStr, Code, Coding, CodeableConcept, HttpUrl, StrictInt
    ] = None

    @validator("value", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, v, values, field):
        return deterimine_choice_type(cls, v, values, field)

    def __str__(self) -> str:
        return f"{self.name}:{self.value}"


class ResourceParameter(AbstractParameter):
    resource: Optional[Resource]


class MultiPartParameter(AbstractParameter):
    part: List[ValueParameter]

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.part:
                if param.name == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return "\n\tpart: \n" + "\n\t".join(" " + str(p) for p in self.part)


class Parameters(Resource):
    resourceType: Literal["Parameters"] = Field("Parameters", const=True)
    parameter: List[Union[ValueParameter, ResourceParameter, MultiPartParameter]]

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.parameter:
                if param.name == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return "Parameters: \n" + "\n".join(" " + str(p) for p in self.parameter)



import logging
from pathlib import Path
from typing import Callable, Generator, Generic, Optional, Sequence, TypeVar, Union

from pydantic import HttpUrl, ValidationError, parse_file_as, parse_obj_as
from fhirkit.elements.elements import CodeableConcept, Coding
from fhirkit.primitive_datatypes import URI, Code
from fhirkit.Bundle import Bundle
from fhirkit.Server import AbstractFHIRServer, ResourceNotFoundError
from fhirkit.TerminologyServer import AbstractFHIRTerminologyServer
from fhirkit.Resource import Resource
from fhirkit.parse import AnyPatientResource, parse_json_as_resource
from tqdm import tqdm

LOGGER = logging.getLogger(__name__)
R = TypeVar("R", bound=Resource)


def traverse(
    path: Path, filter_expr: Optional[Callable[[Path], bool]] = None
) -> Generator[Path, None, None]:
    for childPath in path.iterdir():
        if childPath.is_dir():
            if filter_expr:
                yield from [
                    path
                    for path in traverse(childPath, filter_expr)
                    if filter_expr(path)
                ]
            else:
                yield from traverse(childPath)
        else:
            if filter_expr is None or filter_expr(childPath):
                yield childPath


class SimpleFHIRStore(Generic[R], AbstractFHIRTerminologyServer, AbstractFHIRServer):
    """Simple FHIR Server holds a list of resources in memory. It can be initialised based on bulk export directory."""

    def __init__(
        self, resources: Sequence[R], base_url: Optional[Union[str, HttpUrl]] = None
    ) -> None:
        self._resources = list(resources)
        super().__init__(base_url)

    def iter(self):
        """Dummy method because we can't expect all child classes to be iterable."""
        yield from self._resources

    def __iter__(self):
        return self.iter()

    def __len__(self):
        return len(self._resources)

    def filter(self, expr: Callable[[R], bool]):
        return SimpleFHIRStore(
            [r for r in self.iter() if expr(r)], base_url=self.base_url
        )

    def get_resource(
        self,
        resourceType: Optional[str],
        *,
        id: Optional[str] = None,
        url: Optional[str] = None,
    ):
        if url is not None and self.base_url is not None:
            assert url.startswith(self.base_url), (
                "Can't resolve a resource that is not managed by this server (base URL=%s)"
                % self.base_url,
                url,
            )
            *_, resourceType, id = url.split("/")
        if id is not None:
            assert self._resources is not None
            for r in self._resources:
                if (
                    r.resourceType == resourceType or resourceType is None
                ) and r.id == id:
                    return r
        if id is None and url is None:
            raise RuntimeError(
                "At least an id or url must be given to be able to identify the requested resource"
            )
        t = resourceType or "Resource"
        raise ResourceNotFoundError(f"{t} with id={id} not found.")

    def get_terminology_resource(
        self,
        resourceType: Optional[str],
        *,
        id: Optional[str] = None,
        url: Optional[URI] = None,
    ):

        # ValueSets and CodeSystems have a url that identifies them
        assert (
            resourceType in ("ValueSet", "CodeSystem", "ConceptMap")
            or resourceType is None
        ), "ResourceType should be either %s or %s " % (
            "ValueSet",
            "CodeSystem",
            "ConceptMap",
        )
        try:
            for resource in self._resources:
                if (
                    resource.resourceType in ("ValueSet", "CodeSystem", "ConceptMap"),
                    resource.url == url
                    and url is not None
                    and (resourceType is None or resource.resourceType == resourceType),
                ):
                    return resource
            if id is not None:
                return super().get_resource(resourceType, id=id)
        except ResourceNotFoundError:
            raise
        except Exception:
            LOGGER.warning(
                "Unexpected exception while resolving resource.", exc_info=True
            )
            raise ResourceNotFoundError(
                "Could not resolve resource with resourceType=%s, url=%s, id=%s"
                % (resourceType, url, id)
            )

    def valueset_expand(self, *args, **kwargs):
        return super().valueset_expand(*args, **kwargs)

    def valueset_validate_code(
        self,
        url: URI,
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[URI] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        return super().valueset_validate_code(
            url, code, display, system, coding, codeableConcept
        )

    def _repr_html_(self):
        return (
            """
        <div>
            <style scoped>
                .dataframe tbody tr th:only-of-type {
                    vertical-align: middle;
                }

                .dataframe tbody tr th {
                    vertical-align: top;
                }

                .dataframe thead th {
                    text-align: right;
                }
            </style>
            <table border="1" class="dataframe">
                <thead>
                    <tr style="text-align: right;">
                    <th></th>
                    <th>resourceType</th>
                    <th>id</th>
                    </tr>
                </thead>
            <tbody>
            """
            + "".join(
                [
                    f"<tr><th>{i}</th><td>{r.resourceType}</td><td>{r.id}</td></tr>"
                    for i, r in enumerate(self._resources[:10])
                ]
            )
            + "<tr><th>...</th><td>...</td><td>...</td></tr>"
            if len(self) > 10
            else ""
            + """
                </tbody>
                </table>
                <p>%d resources</p>
                </div>
                """
            % len(self)
        )

    @classmethod
    def bulk_import(cls, path: Path):
        """Load resources from a directory containing NDJSON files with FHIR resources as if they are exported from the FHIR Bulk export API."""
        resources = []
        if isinstance(path, str):
            path = Path(path)
        for fpath in traverse(path, lambda p: p.suffix == ".ndjson"):
            for i, line in enumerate(open(fpath, "r")):
                try:
                    resources.append(parse_json_as_resource(line))
                except ValidationError:
                    LOGGER.warning(
                        "Couldn't parse line %d from '%s'",
                        i,
                        str(fpath.absolute()),
                        exc_info=True,
                    )
        return cls(resources)

    @classmethod
    def load_bundles(cls, path: Path):
        """Load resources from a directory containing JSON files with FHIR Bundle resources per Patients."""
        resources = []
        if isinstance(path, str):
            path = Path(path)
        for fpath in traverse(path, lambda p: p.suffix == ".json"):
            bundle: Bundle = parse_file_as(Bundle, fpath)
            for i, entry in tqdm(enumerate(bundle.entry), desc=str(fpath)):
                if entry.resource is None:
                    LOGGER.warning(
                        "Entry %d in Bundle with path %s has no resoruce.",
                        i,
                        str(fpath.absolute()),
                    )
                    continue
                try:
                    resources.append(
                        parse_obj_as(AnyPatientResource, entry.resource.dict())
                    )
                except ValidationError:
                    LOGGER.warning(
                        "Couldn't load entry %d (fullURL='%s') from '%s'",
                        i,
                        entry.fullUrl,
                        str(fpath.absolute()),
                        exc_info=True,
                    )
        return cls(resources)



from fhirkit.ValueSet import VSCompose, ValueSet


class CompositionValueSet(ValueSet):
    compose: VSCompose

    def expand(self):
        for composition in self.compose.include:
            if len(composition.valueSet) > 0:
                pass
            if len(composition.concept) > 0:
                pass
            if len(composition.filter) > 0:
                pass



from typing import Union


try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import Field, parse_obj_as
from fhirkit import (
    Observation,
    Procedure,
    Condition,
    Immunization,
    DiagnosticReport,
    DocumentReference,
    MedicationAdministration,
    Medication,
    MedicationRequest,
    Encounter,
    CarePlan,
    CareTeam,
    Practitioner,
    Provenance,
    Organization,
    AllergyIntolerance,
    Location,
    PractitionerRole,
    ImagingStudy,
    Patient,
    Device,
    MedicationStatement,
    SupplyDelivery,
)

AnyPatientResource = Annotated[
    Union[
        Observation,
        Procedure,
        Condition,
        Immunization,
        DiagnosticReport,
        DocumentReference,
        MedicationRequest,
        Encounter,
        CarePlan,
        CareTeam,
        Practitioner,
        Organization,
        AllergyIntolerance,
        Location,
        ImagingStudy,
        PractitionerRole,
        Patient,
        MedicationStatement,
        MedicationAdministration,
        Medication,
        Device,
        Provenance,
        SupplyDelivery,
    ],
    Field(discriminator="resourceType"),
]


def parse_json_as_resource(json: str):
    return parse_obj_as(AnyPatientResource, json)



from __future__ import annotations
import logging
from urllib.parse import urlencode

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field

from typing import ClassVar, Optional, Union

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

import requests
from pydantic import Field, HttpUrl, ValidationError, parse_obj_as
from fhirkit.Parameter import Parameters
from fhirkit.TerminologyServer import AbstractFHIRTerminologyServer
from fhirkit.ValueSet import VSExpansion, ValueSet
from fhirkit.primitive_datatypes import URI, Code
from fhirkit.elements import CodeableConcept, Coding
from fhirkit.OperationOutcome import OperationOutcome, OperationOutcomeException


class ExpandedValueset(ValueSet):
    status: Literal["active"] = Field("active")
    expansion: VSExpansion


Response = Annotated[
    Union[Parameters, OperationOutcome], Field(discriminator="resourceType")
]
VSExpansionResponse = Annotated[
    Union[ExpandedValueset, OperationOutcome], Field(discriminator="resourceType")
]


class SCTFHIRTerminologyServer(AbstractFHIRTerminologyServer):
    DEFAULT_URL: ClassVar[
        str
    ] = "https://browser.ihtsdotools.org/snowstorm/snomed-ct/fhir"
    DEFAULT_SERVER: ClassVar[Optional[SCTFHIRTerminologyServer]] = None
    RETRY_COUNT: int = 3
    RETRY_PAUSE: int = 10

    def __init__(self, base_url: Optional[Union[str, HttpUrl]] = None) -> None:
        base_url = base_url or self.DEFAULT_URL
        self._base_url = parse_obj_as(HttpUrl, base_url)

    @classmethod
    def default_server(cls):
        if cls.DEFAULT_SERVER is None:
            cls.DEFAULT_SERVER = SCTFHIRTerminologyServer()
        return cls.DEFAULT_SERVER

    def get_terminology_resource(
        self,
        resourceType: Optional[str],
        *,
        id: Optional[str] = None,
        url: Optional[URI] = None,
    ):
        return super().get_terminology_resource(resourceType, id=id, url=url)

    def get_resource(
        self, resourceType: str, *, id: Optional[str] = None, url: Optional[str] = None
    ):
        path = resourceType
        if id:
            path += "/" + id
        req_url = f"{self.base_url}/{path}"
        if url:
            req_url = url
        headers = {"Accept": "application/json", "Content-type": "application/json"}
        try:
            raw_response = requests.get(req_url, headers=headers)

            if raw_response.status_code != 200:
                logging.warn(
                    "Received an error from the snowstorm server (%s) after sending the following request %s",
                    raw_response.text,
                    req_url,
                )
            response = parse_obj_as(Response, raw_response.json())
        except ValidationError:
            raise ConnectionError(
                f"Received a response that doesn't resemble a FHIR-resource. Please check if the server at {self.base_url} is a valid FHIR-server"
            )
        except:
            raise RuntimeWarning(
                "Failed when calling {endpoint} on {base_url}".format(
                    endpoint=path, base_url=self.base_url
                )
            )

        if isinstance(response, OperationOutcome):
            raise OperationOutcomeException(response)

        return response

    def valueset_expand(self, url: Union[str, HttpUrl], **kwargs):
        path = "ValueSet/$expand"
        page_size = 200
        offset = 0

        more_results_available = True

        while more_results_available:
            query = "url={url}&count={count}&offset={offset}".format(
                url=url, count=page_size, offset=offset
            )
            query += "&" + urlencode(kwargs) if kwargs else ""
            req_url = f"{self.base_url}/{path}?{query}"
            headers = {
                "Accept": "application/json",
            }
            payload = ""

            try:

                raw_response = requests.request(
                    "GET", req_url, data=payload, headers=headers
                )

                if raw_response.status_code != 200:
                    logging.warn(
                        "Received an error from the snowstorm server (%s) after sending the following request %s",
                        raw_response.text,
                        req_url,
                    )
                response = parse_obj_as(VSExpansionResponse, raw_response.json())

                if isinstance(response, OperationOutcome):
                    raise OperationOutcomeException(response)

                yield response

                offset = response.expansion.offset + page_size
                remaining = max(response.expansion.total - offset, 0)

            except ValidationError:
                raise ConnectionError(
                    f"Received a response that doesn't resemble a FHIR-server. Please check if the server at {self.base_url} is a valid FHIR-server. \n Response:\n\n {raw_response.json()}"
                )
            except:
                raise
                raise RuntimeWarning(
                    "Failed when calling {endpoint} on {base_url}".format(
                        endpoint=path, base_url=self.base_url
                    )
                )
            else:

                if remaining % (page_size * 10) == 0:
                    logging.debug(
                        f"{remaining} concepts remaining (total: {response.expansion.total} | page size: {page_size})."
                    )
                more_results_available = remaining > 0

    def valueset_validate_code(
        self,
        url: Union[str, HttpUrl],
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[HttpUrl] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        assert (
            code or coding or codeableConcept
        ), "At least a code, coding or codeableConcept must be given to validate."

        path = "ValueSet/$validate-code"

        parameters = [dict(name="url", valueUri=str(url))]

        if code and system:
            parameters.append(dict(name="code", valueCode=code))
            parameters.append(dict(name="system", valueUri=str(system)))
            if display:
                parameters.append(dict(name="display", valueString=display))
        if coding:
            parameters.append(dict(name="coding", valueCoding=coding))
        if codeableConcept:
            parameters.append(
                dict(name="codeableConcept", valueCodeableConcept=codeableConcept)
            )

        req_url = f"{self.base_url}/{path}"
        headersList = {"Accept": "application/json", "Content-type": "application/json"}
        payload = Parameters(parameter=parameters).json(exclude_none=True)
        try:

            raw_response = requests.request(
                "POST", req_url, data=payload, headers=headersList
            )
            if raw_response.status_code != 200:
                logging.warn(
                    "Received an error from the snowstorm server (%s) after sending the following request %s",
                    raw_response.text,
                    req_url,
                )
            response = parse_obj_as(Response, raw_response.json())
        except ValidationError:
            raise ConnectionError(
                f"Received a response that doesn't resemble a FHIR-server. Please check if the server at {self.base_url} is a valid FHIR-server"
            )
        except:
            raise RuntimeWarning(
                "Failed when calling {endpoint} on {base_url}".format(
                    endpoint=path, base_url=self.base_url
                )
            )

        if isinstance(response, OperationOutcome):
            raise OperationOutcomeException(response)

        return response


def get_default_terminology_server():
    return SCTFHIRTerminologyServer.default_server()


DEFAULT_TERMINOLOGY_SERVER = SCTFHIRTerminologyServer()



from __future__ import annotations

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import (
    AbstractSet,
    Any,
    List,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    Generator,
)
from pydantic import Extra, Field, PrivateAttr
from fhirkit import BaseModel
from fhirkit.primitive_datatypes import URI, Code, Id, Instant, dateTime
from fhirkit.elements import (
    AbstractFHIRServer,
    BackboneElement,
    CodeableConcept,
    ContactDetail,
    Element,
    Identifier,
    Narrative,
    Extension,
    Coding,
    Period,
    Range,
    UsageContext,
)


class Meta(BaseModel):
    versionId: Optional[Id]
    lastUpdated: Optional[Instant]
    source: Optional[URI]
    profile: List[URI] = []
    security: List[Coding] = []
    tag: List[Coding] = []


class Resource(BaseModel):
    _fhir_server: Optional[AbstractFHIRServer] = PrivateAttr(None)

    resourceType: str
    id: Optional[str] = Field(None, repr=False)
    meta: Optional[Meta] = Field(None, repr=False)
    implicitRules: Optional[URI] = Field(None, repr=False)
    language: Optional[Code] = Field(None, repr=False)

    def record(
        self,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_empty: bool = True,
    ):
        tuple_generator = self._iter(
            False,
            by_alias,
            include,
            exclude,
            exclude_unset,
            exclude_defaults,
            exclude_empty,
        )
        return dict(self._assemble_key_recursively(tuple_generator))

    def dict(
        self,
        *,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_empty: bool = False,
    ) -> Dict[str, Any]:
        return dict(
            super()._iter(
                True,
                by_alias,
                include,
                exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_empty=exclude_empty,
                exclude_none=exclude_none,
            )
        )

    def _assemble_key_recursively(
        self, obj: Union[Element, Resource]
    ) -> Generator[Tuple[Tuple[str, ...], Element]]:
        if isinstance(obj, (tuple, list)):
            for key, value in enumerate(obj):
                for childKeys, childValue in self._assemble_key_recursively(value):
                    yield (key, *childKeys), childValue
        elif isinstance(obj, (BackboneElement, Resource, Generator, Period, Range)):
            for key, value in obj:
                for childKeys, childValue in self._assemble_key_recursively(value):
                    yield (key, *childKeys), childValue
        else:
            yield (), obj

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow


class DomainResource(Resource):
    text: Optional[Narrative] = Field(None, repr=False)
    contained: Sequence[Resource] = Field([], repr=False)
    extension: Sequence[Extension] = Field([], repr=False)
    modifierExtension: Sequence[Extension] = Field([], repr=False)

    def _repr_html_(self):
        if self.text:
            return self.text.div
        else:
            return repr(self)


class CanonicalResource(DomainResource):
    url: Optional[URI] = None
    identifier: Sequence[Identifier] = []
    version: Optional[str] = None
    name: Optional[str] = None
    title: Optional[str] = None
    status: Literal["draft", "active", "retired", "unknown"]
    experimental: Optional[bool] = None
    date: Optional[Union[date, dateTime]] = None
    publisher: Optional[str] = None
    contact: Sequence[ContactDetail] = []
    description: Optional[str] = None
    useContext: Sequence[UsageContext] = []
    jurisdiction: Sequence[CodeableConcept] = []
    purpose: Optional[str] = None
    copyright: Optional[str] = None



from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Generator,
    Mapping,
    Optional,
    Tuple,
    Union,
    cast,
)
import pydantic

from pydantic.utils import ROOT_KEY

from fhirkit.choice_type.validators import TYPE_NAME_ALIAS, get_matching_type

IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]
DictStrAny = Dict[str, Any]


class BaseModel(pydantic.BaseModel):
    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    def _iter(
        self,
        to_dict: bool = False,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_empty: bool = True,
    ) -> Generator[Tuple[str, Any], None, None]:
        for k, v in super()._iter(
            to_dict,
            by_alias,
            include,
            exclude,
            exclude_unset,
            exclude_defaults,
            exclude_none,
        ):
            if exclude_empty:
                if v is None:
                    continue
                try:
                    if len(v) == 0:
                        continue
                except TypeError:
                    pass
            if by_alias and k in self.__fields__:
                field = self.__fields__[k]
                if "choice_type" in field.field_info.extra:
                    # notice that 'v' can be a dict when are being called from Model.dict()
                    # this is why we use self.__getattribute to get the original type
                    type_class = get_matching_type(self.__getattribute__(k), field)
                    suffix = TYPE_NAME_ALIAS.get(
                        type_class.__name__, type_class.__name__
                    )
                    suffix = suffix[:1].title() + suffix[1:]
                    k = k + suffix
            yield k, v

    def dict(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_empty: bool = True,
    ) -> DictStrAny:
        return dict(
            self._iter(
                to_dict=True,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                exclude_empty=exclude_empty,
            )
        )

    def json(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        by_alias: bool = True,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
        exclude_empty: bool = True,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        encoder = cast(Callable[[Any], Any], encoder or self.__json_encoder__)

        # We don't directly call `self.dict()`, which does exactly this with `to_dict=True`
        # because we want to be able to keep raw `BaseModel` instances and not as `dict`.
        # This allows users to write custom JSON encoders for given `BaseModel` classes.
        data = dict(
            self._iter(
                to_dict=models_as_dict,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                exclude_empty=exclude_empty,
            )
        )
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        return self.__config__.json_dumps(data, default=encoder, **dumps_kwargs)



from __future__ import annotations
from functools import total_ordering
import itertools
from typing import Any, ForwardRef, Optional, Sequence, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from pydantic import AnyUrl, Field
from fhirkit import BaseModel
from fhirkit.primitive_datatypes import URI, Code, XHTML, dateTime
from fhirkit.Server import AbstractFHIRServer


class Element(BaseModel):
    id: Optional[str]
    extension: Sequence[Extension] = Field([], repr=False)


class Attachment(Element):
    contentType: Optional[Code]
    language: Optional[Code]
    # TODO how to type base64 encode bytes


class Narrative(Element):
    status: Literal["generated", "extensions", "additional", "required"]
    div: XHTML


class Extension(Element):
    url: URI
    value: Optional[Any]


class BackboneElement(Element):
    modifierExtension: Sequence[Extension] = Field([], repr=False)


Narrative.update_forward_refs()
BackboneElement.update_forward_refs()
Element.update_forward_refs()


class AbstractCoding(Element):
    """FHIR Terminology based model for concepts"""

    display: Optional[str]
    code: str
    system: Optional[URI]
    version: Optional[str]

    class Config:
        allow_mutation = False


@total_ordering
class Coding(AbstractCoding):
    def __repr__(self) -> str:
        if self.display:
            return f'"{self.display}" {self.system}|{self.code}'
        else:
            return f"{self.system}|{self.code}"

    def __str__(self) -> str:
        return self.display or f"{self.system}|{self.code}"

    def fsh(self, include_display: bool = False, include_version: bool = True):
        token = self.system
        if self.version is not None and include_version:
            token += "|" + self.version
        token += "#" + self.code
        if self.display is not None and include_display:
            token += ' "{self.display}"'
        return token

    def __eq__(self, other: AbstractCoding) -> bool:
        if isinstance(other, AbstractCoding):
            return self.system == other.system and self.code == other.code
        return False

    def __lt__(self, other: AbstractCoding) -> bool:
        return (self.system, self.code) < (other.system, other.code)

    def __hash__(self):
        return hash((self.system, self.code))


CodeableConcept = ForwardRef("CodeableConcept")


class CodeableConcept(BaseModel):
    """FHIR Terminology based model for CodeableConcepts"""

    text: Optional[str] = None
    coding: Sequence[Coding] = Field([])
    active: Optional[bool] = None

    def __str__(self) -> str:
        return self.text

    def __eq__(self, other: Union[CodeableConcept, Coding]) -> bool:

        if isinstance(other, Coding):
            result = any(other == c for c in self.coding)
            return result
        elif isinstance(other, CodeableConcept):
            return any(
                c1 == c2 for c1, c2 in itertools.product(self.coding, other.coding)
            )
        else:
            return False


CodeableConcept.update_forward_refs()


class Period(Element):
    start: Optional[dateTime]
    end: Optional[dateTime]


class Reference(Element):
    reference: Optional[str]
    type: Optional[URI]
    identifier: Optional[Identifier]
    display: Optional[str]

    def __repr__(self) -> str:
        if self.display is not None:
            return self.display
        if self.type is not None and self.reference is not None:
            return self.type + "/" + self.reference
        return super().__repr__()

    def resolve(self, store: AbstractFHIRServer):
        if self.reference:
            return store[self.reference.replace("urn:uuid:", "")]


class Identifier(Element):
    use: Optional[Code]
    type: Optional[CodeableConcept]
    system: Optional[URI]
    value: Optional[str]
    period: Optional[Period]
    assigner: Optional[Reference]


Identifier.update_forward_refs()
Reference.update_forward_refs()


class Quantity(Element):
    value: float
    comparator: Optional[Literal["<", "<=", ">=", ">"]]
    unit: str
    system: Optional[AnyUrl]
    code: Optional[Code]

    def __str__(self) -> str:
        return f"{self.value} {self.unit}"

    def unit_as_coding(self):
        return Coding(display=self.unit, code=self.code, system=self.system)

    def __repr__(self) -> str:
        rep = f"{self.value} {self.unit}"
        if self.comparator:
            return self.comparator + " " + rep
        return rep


class SimpleQuantity(Quantity):
    comparator: None = Field(..., const=True)


# TODO specify these derived quantities further
class Age(Quantity):
    pass


class Duration(Quantity):
    pass


class Range(Element):
    low: Optional[SimpleQuantity]
    high: Optional[SimpleQuantity]


class Ratio(Element):
    numerator: Optional[Quantity] = None
    denominator: Optional[Quantity] = None


class ContactPoint(Element):
    system: Optional[
        Literal["phone", "fax", "email", "pager", "url", "sms", "other"]
    ] = None
    value: Optional[str] = None
    use: Optional[Literal["home", "work", "temp", "old", "mobile"]] = None
    period: Optional[Period] = None


class ContactDetail(Element):
    name: Optional[str]
    telecom: Sequence[ContactPoint] = []



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import Optional
from pydantic import Field


from fhirkit.Resource import DomainResource
from .elements.elements import CodeableConcept, Reference


class Encounter(DomainResource):
    resourceType: Literal["Encounter"] = Field("Encounter", const=True)
    subject: Optional[Reference] = None
    reasonCode: Optional[CodeableConcept] = None
    reasonReference: Optional[Reference] = None



from fhirkit import Observation
import pytest_check as check
import json

from fhirkit.elements.elements import Quantity


def test_parse_observation():
    obs = Observation.parse_file("./test/observation.json")
    obs_dict = obs.dict()
    check.is_in("value", obs_dict)
    check.is_instance(obs.value, Quantity)
    check.is_not_in("valueQuantity", obs_dict)
    reconstructed = json.loads(obs.json())
    check.is_in("valueQuantity", reconstructed)



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field


from fhirkit.Resource import DomainResource


class AllergyIntolerance(DomainResource):
    resourceType: Literal["AllergyIntolerance"] = Field(
        "AllergyIntolerance", const=True
    )



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field


from fhirkit.Resource import DomainResource


class Organization(DomainResource):
    resourceType: Literal["Organization"] = Field("Organization", const=True)



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import Optional, Sequence
from pydantic import Field
from fhirkit.Resource import DomainResource
from fhirkit.elements import BackboneElement, CodeableConcept
from fhirkit.primitive_datatypes import Code


class OperationOutcomeIssue(BackboneElement):
    severity: Literal["fatal", "error", "warning", "information"]
    code: Code  # TODO add logic that can bind this to valueset
    details: Optional[CodeableConcept] = None
    diagnostics: Optional[str] = None
    location: Sequence[str] = []
    expression: Sequence[str] = []

    def __str__(self):
        txt = f"[{self.severity}](issue type={self.code}) "
        if self.diagnostics:
            txt += self.diagnostics
        return txt


class OperationOutcome(DomainResource):
    resourceType: Literal["OperationOutcome"] = Field("OperationOutcome", const=True)
    issue: Sequence[OperationOutcomeIssue] = []

    def __str__(self) -> str:
        return "OperationOutcome: \n" + "\n\n".join(str(issue) for issue in self.issue)


class OperationOutcomeException(Exception):
    def __init__(self, resource: OperationOutcome) -> None:
        self.resource = resource
        super().__init__(str(resource))



from datetime import datetime  # this is important to have at the top

from typing import List, Optional, Sequence, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


from pydantic import Field, validator
from fhirkit.choice_type import deterimine_choice_type, ChoiceType
from fhirkit.Resource import DomainResource
from fhirkit.primitive_datatypes import Code, time
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Quantity,
    Range,
    Ratio,
    Reference,
    BackboneElement,
)

ValueType = Union[
    str, Quantity, int, CodeableConcept, bool, Range, Ratio, time, datetime, Period
]


class ObservationComponent(BackboneElement):
    code: CodeableConcept
    valueString: Optional[str] = Field(None, exclude=True)
    valueQuantity: Optional[Quantity] = Field(None, exclude=True)
    valueInteger: Optional[int] = Field(None, exclude=True)
    valueCodeableConcept: Optional[CodeableConcept] = Field(None, exclude=True)
    valueBoolean: Optional[bool] = Field(None, exclude=True)
    valueRatio: Optional[Ratio] = Field(None, exclude=True)
    valueTime: Optional[time] = Field(None, exclude=True)
    valueDateTime: Optional[datetime] = Field(None, exclude=True)
    valuePeriod: Optional[Period] = Field(None, exclude=True)
    value: Optional[ValueType] = ChoiceType(None)

    @validator("value", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, v, values, field):
        return deterimine_choice_type(cls, v, values, field)


class Observation(DomainResource):

    resourceType: Literal["Observation"] = Field("Observation", const=True)
    identifier: Sequence[Identifier] = Field([], repr=True)
    status: Literal[
        "registered",
        "preliminary",
        "final",
        "amended",
        "corrected",
        "cancelled",
        "enterred-in-error",
        "unknown",
    ] = Field("final", repr=True)
    category: Sequence[CodeableConcept] = Field([], repr=True)

    code: CodeableConcept = Field(..., repr=True)
    valueString: Optional[str] = Field(None, exclude=True)
    valueQuantity: Optional[Quantity] = Field(None, exclude=True)
    valueInteger: Optional[int] = Field(None, exclude=True)
    valueCodeableConcept: Optional[CodeableConcept] = Field(None, exclude=True)
    valueBoolean: Optional[bool] = Field(None, exclude=True)
    valueRatio: Optional[Ratio] = Field(None, exclude=True)
    valueTime: Optional[time] = Field(None, exclude=True)
    valueDateTime: Optional[datetime] = Field(None, exclude=True)
    valuePeriod: Optional[Period] = Field(None, exclude=True)
    value: Optional[ValueType] = ChoiceType(None)

    subject: Optional[Reference]
    encounter: Optional[Reference]

    method: Optional[Code] = Field(None, repr=False)
    derivedFrom: Optional[Reference]

    component: List[ObservationComponent] = Field([], repr=True)

    effectivePeriod: Optional[Period] = Field(None, exclude=True)
    effectiveDateTime: Optional[datetime] = Field(None, exclude=True)
    effective: Union[datetime, Period] = ChoiceType(None)

    @validator("value", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, v, values, field):
        return deterimine_choice_type(cls, v, values, field)

    @validator("effective", pre=True, always=True)
    def validate_effective(cls, v, values, field):
        return deterimine_choice_type(cls, v, values, field)



from __future__ import annotations
from email.generator import Generator
from typing import Any, ClassVar, Optional, Sequence, Set, TypeVar, Union

from fhirkit.choice_type import ChoiceType

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from pydantic import AnyUrl, Field, StrictBool, StrictStr, validator
from fhirkit.Resource import CanonicalResource
from fhirkit.choice_type import deterimine_choice_type
from fhirkit.elements import (
    BackboneElement,
    CodeableConcept,
    Coding,
    Identifier,
    UsageContext,
)
from fhirkit.primitive_datatypes import URI, Code, dateTime, date
from fhirkit.metadata_types import ContactDetail


class CodeLookupError(KeyError):
    pass


class CSConceptDesignation(BackboneElement):
    language: Optional[Code]
    use: Optional[Coding]
    value: str


class CSConceptProperty(BackboneElement):
    code: Code
    valueBoolean: Optional[StrictBool] = Field(None, exclude=True)
    valueString: Optional[StrictStr] = Field(None, exclude=True)
    valueCode: Optional[Code] = Field(None, exclude=True)
    valueCoding: Optional[Coding] = Field(None, exclude=True)
    valueDateTime: Optional[dateTime] = Field(None, exclude=True)
    valueDecimal: Optional[float] = Field(None, exclude=True)
    value: Union[StrictBool, StrictStr, Code, Coding, dateTime, float] = ChoiceType(
        None
    )

    @validator("value", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, v, values, field):
        return deterimine_choice_type(cls, v, values, field)

    def __str__(self) -> str:
        return str(self.code) + ": " + str(self.value)


class CSConcept(BackboneElement):
    code: Code
    display: Optional[str]
    definition: Optional[str]
    designation: Sequence[CSConceptDesignation] = []
    property: Sequence[CSConceptProperty] = []
    concept: Sequence[CSConcept] = []

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.property:
                if param.code == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return (
            f"#{self.code} '{self.display}'"
            + "\nproperties: \n"
            + "\n".join(" " + str(p) for p in self.property if p.value is not None)
        )


class CSConceptLookup(CSConcept):
    name: str


C = TypeVar("C", bound=CSConcept)


def traverse_concepts(
    s: Sequence[C],
) -> Generator[C, None, None]:
    for c in s:
        yield c
        if len(c.concept) > 0:
            yield from traverse_concepts(c.concept)


class CodeSystem(CanonicalResource):
    resourceType = Field("CodeSystem", const=True)
    url: Optional[URI]
    identifier: Sequence[Identifier] = []
    version: Optional[str]
    name: Optional[str]
    title: Optional[str]
    status: Literal["draft", "active", "retired", "unknown"]
    experimental: Optional[bool]
    date: Optional[Union[date, dateTime]]
    publisher: Optional[str]
    contact: Sequence[ContactDetail] = []
    description: Optional[str]
    useContext: Sequence[UsageContext] = []
    jurisdiction: Sequence[CodeableConcept] = []
    purpose: Optional[str]
    copyright: Optional[str]
    caseSensitive: Optional[bool]
    valueSet: Optional[URI]
    hierarchyMeaning: Optional[
        Literal["grouped-by", "is-a", "part-of", "classified-with"]
    ]
    compositional: Optional[bool]
    versionNeeded: Optional[bool]
    content: Literal["not-present", "example", "fragment", "complete", "supplement"]
    supplements: Optional[URI]
    count: Optional[int]
    concept: Sequence[CSConcept] = []

    def lookup(
        self,
        code: Optional[Code] = None,
        coding: Optional[Coding] = None,
        date: Optional[dateTime] = None,
        displayLanguage: Optional[Code] = None,
        property: Optional[Code] = None,
    ):
        if len(self.concept) == 0:
            raise RuntimeWarning(
                "No strategy to lookup codes or concepts if concepts are not explicitly available under 'CodeSystem.concept'"
            )

        if self.name is None:
            raise RuntimeWarning("The CodeSystem has no name to use as display name")
        assert (
            code is not None or coding is not None
        ), "At least a code or coding is needed to lookup."
        for concept in traverse_concepts(self.concept):

            if (concept.code == code) or (
                coding
                and concept.code == coding.code
                and concept.display == coding.display
            ):
                return CSConceptLookup(
                    name=self.name,
                    **concept.dict(
                        include={
                            "code",
                            "designation",
                            "property",
                            "display",
                            "version",
                        }
                    ),
                )

        raise CodeLookupError(
            f"No concept found in CodeSystem for given code/coding. (code={code}, coding={coding})"
        )

    def __iter__(self):
        yield from traverse_concepts(self.concept)

    def __len__(self):
        if self.count:
            return self.count
        if self.content == "complete":
            self.count = len(self)
        raise ValueError(
            "CodeSystem.count is None and we have not strategy to estimate the number of concepts when CodeSystem.content != 'complete'"
        )

    def __getitem__(self, key: Code | Coding) -> CSConceptLookup:

        if isinstance(key, (Code, str)):
            return self.lookup(code=key)
        elif isinstance(key, Coding):
            return self.lookup(coding=key)
        raise KeyError("Key should be a Code or a Coding but received " + key)


CodeSystem.update_forward_refs()
CSConcept.update_forward_refs()



from datetime import date, datetime

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import Iterable, List, Optional, Sequence, Union
from pydantic import Field
from fhirkit import BaseModel
from fhirkit.primitive_datatypes import URI, dateTime
from fhirkit.elements import (
    BackboneElement,
    CodeableConcept,
    Coding,
    Narrative,
    UsageContext,
)
from fhirkit.Resource import CanonicalResource


class VSDesignation(BaseModel):
    language: Optional[str]
    use: Optional[Coding]
    value: str


class VSConcept(BaseModel):
    code: str
    display: Optional[str]
    designation: List[VSDesignation] = Field(default=[])


class VSFilter(BackboneElement):
    property: str
    op: Literal[
        "=",
        "is-a",
        "descendent-of",
        "is-not-a",
        "regex",
        "in",
        "not-in",
        "generalizes",
        "child-of",
        "descendent-leaf",
        "exists",
    ]
    value: str


class VSInclude(BackboneElement):
    system: Optional[URI] = None
    version: Optional[str] = None
    concept: Sequence[VSConcept] = Field(default=[])
    filter: Sequence[VSFilter] = Field(default=[])
    valueSet: Sequence[URI] = Field(default=[])


class VSCompose(BaseModel):
    include: List[VSInclude] = []
    exclude: List[VSInclude] = []
    property: Sequence[str] = []
    lockedDate: Optional[date]
    inactive: Optional[bool]


class VSCodingWithDesignation(Coding):
    designation: List[VSDesignation] = Field(default=[])


class VSExpansion(BaseModel):
    offset: Optional[int]
    total: Optional[int]
    contains: List[VSCodingWithDesignation] = []
    identifier: Optional[URI] = None
    timestamp: dateTime = Field(default_factory=datetime.now)


class ValueSet(CanonicalResource):
    resourceType: Literal["ValueSet"] = Field("ValueSet", const=True)
    url: Optional[URI]
    name: Optional[str]
    compose: Optional[VSCompose]
    expansion: Optional[VSExpansion]
    useContext: Sequence[UsageContext] = Field([], repr=True)

    def __iter__(self):
        if not self.has_expanded:
            self.expand()
        for coding in self.expansion.contains:
            yield coding

    def __len__(self):
        if not self.has_expanded:
            self.expand()
            assert (
                self.has_expanded
            ), "ValueSet has no expansion even after running `self.expand()`."
        return self.expansion.total or len(self.expansion.contains)

    def __contains__(self, item: Union[Coding, CodeableConcept]) -> bool:
        if not isinstance(item, (Coding, CodeableConcept)):
            return False
        return self.validate_code(item)

    @property
    def has_expanded(self):
        return self.expansion is not None

    def expand(self):
        """Override this method to implement expansion logic.
        This method should fill ValueSet.expansion.contains with concepts.

        Implementing this method enables you to iterate over the ValueSet in a for-loop.
        ```python
        class MyCustomValueSet(ValueSet)

            def expand(self):
                # some expansion logic

        vs_example = MyCustomValueSet()
        for coding in vs:
            print(coding)
        "
        """
        raise NotImplementedError()

    def validate_code(self, code: Union[Coding, CodeableConcept]):
        raise NotImplementedError()

    def init_expansion(self):
        self.expansion = VSExpansion()

    def append(
        self,
        code: VSCodingWithDesignation,
        extend_compose: bool = True,
        init_expansion_if_none: bool = True,
    ):
        if extend_compose:
            self.compose.include.append(VSInclude(system=code.system, concept=[code]))

        if not self.has_expanded and init_expansion_if_none:
            self.init_expansion()

        assert (
            self.expansion is not None
        ), "`self.expansion` is None after initialisation with `self.init_expansion`"
        self.expansion.contains.append(code)

    def extend(
        self,
        codes: Iterable[VSCodingWithDesignation],
        extend_compose: bool = True,
        auto_init_expansion: bool = True,
    ):
        if extend_compose:
            first_code, *_ = iter(codes)
            self.compose.include.append(
                VSInclude(system=first_code.system, concept=list(codes))
            )

        if not self.has_expanded and auto_init_expansion:
            self.init_expansion()
        assert (
            self.expansion is not None
        ), "`self.expansion` is None after initialisation with `self.init_expansion`"
        self.expansion.contains.extend(codes)


class SimpleValueSet(ValueSet):
    status: Literal["active"] = Field("active", const=True)

    def __init__(self, *args: VSCodingWithDesignation, **kwargs):
        if len(args) > 0:

            assert (
                "expansion" not in kwargs
            ), "When passing an iterable with concepts, `expansion` should be None."
            super().__init__(
                expansion=VSExpansion(
                    contains=[c.dict() for c in args], total=len(args)
                ),
                text=Narrative(
                    status="generated",
                    div="""
                <div>
                    <style scoped>
                        .dataframe tbody tr th:only-of-type {
                            vertical-align: middle;
                        }

                        .dataframe tbody tr th {
                            vertical-align: top;
                        }

                        .dataframe thead th {
                            text-align: right;
                        }
                    </style>
                    <table border="1" class="dataframe">
                        <thead>
                            <tr style="text-align: right;">
                            <th>code</th>
                            <th>display</th>
                            <th>system</th>
                            <th>version</th>
                            </tr>
                        </thead>
                        <tbody>"""
                    + "".join(
                        [
                            f"<tr><th>{c.code}</th><td>{c.display}</td><td>{c.system}</td><td>{c.version}</td></tr>"
                            for c in args
                        ]
                    )
                    + """
                        </tbody>
                    </table>
                </div>""",
                ),
                **kwargs,
            )
        else:
            super().__init__(**kwargs)

    def expand(self):
        raise UserWarning(
            "SimpleValueSet is already expanded at construction time. So it doesn't make sense to explicitly ask for expansion."
        )

    def validate_code(self, code: Union[Coding, CodeableConcept]):
        if isinstance(code, CodeableConcept):
            return any(c in self for c in code.coding)
        elif isinstance(code, Coding):
            return any(c == code for c in self)
        else:
            return False


ValueSet.update_forward_refs()



from datetime import datetime
from typing import Optional, Sequence, Union

from pydantic import Field, PositiveInt, validator
from fhirkit.elements import (
    BackboneElement,
    CodeableConcept,
    Duration,
    Element,
    Period,
    Range,
)
from fhirkit.primitive_datatypes import decimal, Code
from fhirkit.choice_type import deterimine_choice_type, ChoiceType


class TimingRepeat(Element):
    boundsDuration: Optional[Duration] = Field(None, exclude=True)
    boundsRange: Optional[Range] = Field(None, exclude=True)
    boundsPeriod: Optional[Period] = Field(None, exclude=True)
    bounds: Optional[Union[Duration, Range, Period]] = ChoiceType(None)

    @validator("bounds", pre=True, always=True, allow_reuse=True)
    def validate_bounds(cls, value, values, field):
        return deterimine_choice_type(cls, value, values, field)

    count: Optional[PositiveInt] = None
    countMax: Optional[PositiveInt] = None
    duration: Optional[decimal] = None
    durationMax: Optional[decimal] = None
    durationUnit: Optional[Code] = None
    frequency: Optional[PositiveInt] = None
    frequencyMax: Optional[PositiveInt] = None
    period: Optional[decimal] = None
    periodMax: Optional[decimal] = None
    periodUnit: Optional[Code] = None
    dayOfWeek: Sequence[Code] = []
    timeOfDay: Sequence[Code] = []
    when: Sequence[Code] = []


TimingRepeat.update_forward_refs()


class Timing(BackboneElement):
    event: Sequence[datetime] = []
    repeat: Optional[TimingRepeat] = None
    code: Optional[CodeableConcept] = None


Timing.update_forward_refs()



from __future__ import annotations
from typing import List, Optional, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import PrivateAttr, root_validator, Field
from fhirkit.ValueSet import VSCompose, VSFilter, VSInclude, ValueSet
from fhirkit.primitive_datatypes import URI
from fhirkit.elements import CodeableConcept, Coding, Narrative
from fhirkit.snomed.consts import SCT_URI
from fhirkit.snomed.terminology import (
    SCTFHIRTerminologyServer,
    get_default_terminology_server,
)


class SCTDescendantsFilter(VSFilter):
    property: Literal["concept"]
    op: Literal["is-a"]
    value: str

    def to_url_query(self):
        return f"fhir_vs=isa/{self.value}"

    def __repr__(self) -> str:
        return f"concept is-a {self.value}"


class SCTECLFilter(VSFilter):
    property: Literal["constraint"]
    op: Literal["="]
    value: str

    def to_url_query(self):
        return f"fhir_vs=ecl/{self.value}"

    def __repr__(self) -> str:
        return f"constraint = {self.value}"


class SCTImplicitCompose(VSCompose):
    include: Tuple[SCTImplicitInclude]


class SCTImplicitValueSet(ValueSet):
    _fhir_server: SCTFHIRTerminologyServer = PrivateAttr(
        default_factory=get_default_terminology_server
    )
    status: Literal["draft", "active", "retired", "unknown"] = "active"
    url: Optional[URI]
    compose: Optional[SCTImplicitCompose]

    @root_validator
    def check_either_url_or_compose(cls, values):
        url, compose = values.get("url"), values.get("compose")
        if url is None and compose is None:
            raise ValueError(
                "Either a 'url' or a 'compose' section should be specified to have a valid ValueSet."
            )
        return values

    def ensure_fhir_server(self):
        if self._fhir_server is None:
            raise RuntimeWarning(
                "Can't expand an implicit SNOMED-CT without a FHIR server."
            )
        return True

    def equivalent_url(self):
        if self.url:
            return self.url
        else:
            assert (
                self.compose is not None
            ), "If no url is specified at least a compose section should be available"
            inclusion = self.compose.include[0]
            filter_rule = inclusion.filter[0]
            if filter_rule.property == "constraint" and filter_rule.op == "=":
                url = inclusion.version or inclusion.system or SCT_URI
                url += "?fhir_vs=ecl/" + inclusion.filter[0].value
            elif filter_rule.property == "concept" and filter_rule.op == "is-a":
                url = inclusion.version or inclusion.system or SCT_URI
                url += "?fhir_vs=isa/" + inclusion.filter[0].value
            else:
                raise RuntimeError(
                    f"Unexepected filter {inclusion.filter[0].json()} in inclusion criterium: {inclusion.json()}"
                )
            return url

    def expand(self, **kwargs):
        self.ensure_fhir_server()
        url = self.equivalent_url()
        self.init_expansion()
        for expanded_vs in self._fhir_server.valueset_expand(url, **kwargs):
            self.extend(expanded_vs.expansion.contains, extend_compose=False)

        self.build_narrative()

    def build_narrative(self):
        self.text = Narrative(
            status="generated",
            div="""
                <div>
                    <style scoped>
                        .dataframe tbody tr th:only-of-type {
                            vertical-align: middle;
                        }

                        .dataframe tbody tr th {
                            vertical-align: top;
                        }

                        .dataframe thead th {
                            text-align: right;
                        }
                    </style>
                    <table border="1" class="dataframe">
                        <thead>
                            <tr style="text-align: right;">
                            <th>code</th>
                            <th>display</th>
                            <th>system</th>
                            <th>version</th>
                            </tr>
                        </thead>
                        <tbody>"""
            + "".join(
                [
                    f"<tr><th>{c.code}</th><td>{c.display}</td><td>{c.system}</td><td>{c.version}</td></tr>"
                    for c in self
                ]
            )
            + """
                        </tbody>
                    </table>
                </div>
            """,
        )

    def validate_code(self, code: Union[Coding, CodeableConcept]):
        self.ensure_fhir_server()
        url = self.equivalent_url()
        assert isinstance(
            code, (Coding, CodeableConcept)
        ), "code should be a Coding or CodeableConcept in order to be able to validate it against a ValueSet"
        if self.has_expanded:
            for member_code in self:
                if code == member_code:
                    return True
            return False
        if isinstance(code, Coding):
            response = self._fhir_server.valueset_validate_code(url, coding=code)
        else:
            response = self._fhir_server.valueset_validate_code(
                url, codeableConcept=code
            )
        return response.result


class SCTImplicitInclude(VSInclude):
    system: URI = Field("http://snomed.info/sct")
    concept: List = Field([], const=True)
    valueSet: List = Field([], const=True)
    filter: Tuple[Union[SCTDescendantsFilter, SCTECLFilter]]

    def equivalent_url(self):
        if self.version is not None:
            return self.version + "?" + self.filter[0].to_url_query()
        return self.system + "?" + self.filter[0].to_url_query()

    def to_valueset(self) -> SCTImplicitValueSet:
        return SCTImplicitValueSet(url=self.equivalent_url())

    def __repr__(self):
        return repr(self.filter[0])


SCTImplicitCompose.update_forward_refs()
SCTImplicitValueSet.update_forward_refs()



from typing import AbstractSet, Any, Callable, Mapping, Optional, Union
from pydantic.fields import FieldInfo, Undefined

IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]
NoArgAnyCallable = Callable[[], Any]


def ChoiceType(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    title: str = None,
    description: str = None,
    exclude: Union["AbstractSetIntStr", "MappingIntStrAny", Any] = None,
    include: Union["AbstractSetIntStr", "MappingIntStrAny", Any] = None,
    const: bool = None,
    gt: float = None,
    ge: float = None,
    lt: float = None,
    le: float = None,
    multiple_of: float = None,
    max_digits: int = None,
    decimal_places: int = None,
    min_items: int = None,
    max_items: int = None,
    unique_items: bool = None,
    min_length: int = None,
    max_length: int = None,
    allow_mutation: bool = True,
    regex: str = None,
    discriminator: str = None,
    repr: bool = True,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field, either for the model schema or complex validation. Some arguments
    apply only to number fields (``int``, ``float``, ``Decimal``) and some apply only to ``str``.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclude: exclude this field while dumping.
      Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param include: include this field while dumping.
      Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param const: this field is required and *must* take it's default value
    :param gt: only applies to numbers, requires the field to be "greater than". The schema
      will have an ``exclusiveMinimum`` validation keyword
    :param ge: only applies to numbers, requires the field to be "greater than or equal to". The
      schema will have a ``minimum`` validation keyword
    :param lt: only applies to numbers, requires the field to be "less than". The schema
      will have an ``exclusiveMaximum`` validation keyword
    :param le: only applies to numbers, requires the field to be "less than or equal to". The
      schema will have a ``maximum`` validation keyword
    :param multiple_of: only applies to numbers, requires the field to be "a multiple of". The
      schema will have a ``multipleOf`` validation keyword
    :param max_digits: only applies to Decimals, requires the field to have a maximum number
      of digits within the decimal. It does not include a zero before the decimal point or trailing decimal zeroes.
    :param decimal_places: only applies to Decimals, requires the field to have at most a number of decimal places
      allowed. It does not include trailing decimal zeroes.
    :param min_items: only applies to lists, requires the field to have a minimum number of
      elements. The schema will have a ``minItems`` validation keyword
    :param max_items: only applies to lists, requires the field to have a maximum number of
      elements. The schema will have a ``maxItems`` validation keyword
    :param max_items: only applies to lists, requires the field not to have duplicated
      elements. The schema will have a ``uniqueItems`` validation keyword
    :param min_length: only applies to strings, requires the field to have a minimum length. The
      schema will have a ``maximum`` validation keyword
    :param max_length: only applies to strings, requires the field to have a maximum length. The
      schema will have a ``maxLength`` validation keyword
    :param allow_mutation: a boolean which defaults to True. When False, the field raises a TypeError if the field is
      assigned on an instance.  The BaseModel Config must set validate_assignment to True
    :param regex: only applies to strings, requires the field match against a regular expression
      pattern string. The schema will have a ``pattern`` validation keyword
    :param discriminator: only useful with a (discriminated a.k.a. tagged) `Union` of sub models with a common field.
      The `discriminator` is the name of this common field to shorten validation and improve generated schema
    :param repr: show this field in the representation
    :param **extra: any additional keyword arguments will be added as is to the schema
    """
    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=None,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        discriminator=discriminator,
        repr=repr,
        choice_type=True,
        **extra,
    )
    field_info._validate()
    return field_info



from setuptools import setup, find_packages

classifiers = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Healthcare Industry",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
]
setup(
    name="FHIRkit",
    version="0.0.23",
    description="Toolkit to handle FHIR Resources in a more efficient, pythonic way.",
    long_description=open("README.md").read(),
    url="",
    author="",
    author_email="",
    license="MIT",
    classifiers=classifiers,
    keywords="",
    packages=find_packages(),
    install_requires=["pydantic", "requests", "tqdm"],
)



from typing import Optional, Union

from pydantic import Field, validator
from fhirkit.choice_type.choice_type import ChoiceType
from fhirkit.elements import (
    CodeableConcept,
    Coding,
    Element,
    Quantity,
    Range,
    Reference,
)

from fhirkit.choice_type import deterimine_choice_type


class UsageContext(Element):
    code: Coding
    valueCodeableConcept: Optional[CodeableConcept] = Field(None, exclude=True)
    valueQuantity: Optional[Quantity] = Field(None, exclude=True)
    valueRange: Optional[Range] = Field(None, exclude=True)
    valueReference: Optional[Reference] = Field(None, exclude=True)
    value: Union[CodeableConcept, Quantity, Range, Reference] = ChoiceType(None)

    @validator("value", pre=True, always=True, allow_reuse=True)
    def validate_value(cls, value, values, field):
        return deterimine_choice_type(cls, value, values, field)


UsageContext.update_forward_refs()



from pydantic import Field, HttpUrl
from fhirkit.elements import CodeableConcept, Coding
from .ValueSet import (
    SCTDescendantsFilter,
    SCTImplicitCompose,
    SCTImplicitInclude,
    SCTImplicitValueSet,
)


class SCTCoding(Coding):
    system: HttpUrl = Field(default="http://snomed.info/sct")

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], str):
            code, display, *_ = (
                x.strip() for x in args[0].split("|")
            )  # break string in pieces
            super().__init__(code=code, display=display, **kwargs)
        else:
            super().__init__(*args, **kwargs)  # business as usual, Pydantic takes over

    def descendants(self):
        return SCTImplicitValueSet(
            url=f"{self.system}?fhir_vs=isa/{self.code}",
            status="active",
            compose=SCTImplicitCompose(
                include=[
                    SCTImplicitInclude(
                        system=self.system,
                        filter=[
                            SCTDescendantsFilter(
                                property="concept", op="is-a", value=self.code
                            )
                        ],
                    )
                ]
            ),
        )


class SCTConcept(CodeableConcept):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], str):
            sct_coding = SCTCoding(args[0])
            super().__init__(text=sct_coding.display, coding=[sct_coding], **kwargs)
        else:
            super().__init__(*args, **kwargs)  # business as usual, Pydantic takes over



from typing import Any, Dict
from pydantic.fields import ModelField

TYPE_NAME_ALIAS = {
    "str": "string",
    "datetime": "dateTime",
    "float": "decimal",
    "int": "integer",
    "bool": "Boolean",
}


def get_matching_type(value: Any, field: ModelField):
    for type_class in field.type_.__args__:
        if isinstance(value, type_class):
            return type_class
    raise TypeError(f"{value} doesn't match any of the specified types ", field.type_)


def deterimine_choice_type(cls, v: Any, values: Dict[str, Any], field: ModelField):
    """Automatically assign the value from one of the choice type alias fields to current field."""
    if v is not None:
        return v

    non_null_values = list(
        filter(
            lambda t: t[0].startswith(field.name) and t[1] is not None,
            values.items(),
        )
    )
    if len(non_null_values) == 0:
        return
    elif len(non_null_values) > 1:
        raise ValueError(
            f"{cls.__class__.__name__}.{field.name}[x] can only have one value."
        )
    v = non_null_values[0][1]
    return v
